import os
from .to_sql_interface import ToSQLInterface


def validate_select_table_name(database_object_name: str) -> None:
    if os.getenv("environment_name") not in ("prod1", "dvlp1") and not database_object_name.endswith("_view"):  # noqa501
        raise Exception(
            f"View name must end with '_view' in this environment (got {database_object_name})")  # noqa501


def validate_none_select_table_name(database_object_name: str) -> None:
    if os.getenv("environment_name") not in ("prod1", "dvlp1") and not database_object_name.endswith("_table"):  # noqa501
        raise Exception(f"Table name must end with '_table' in this environment  (got {database_object_name})")  # noqa501


def process_insert_data_json(data_json):
    columns = []
    values = []

    for key, value in data_json.items():
        if isinstance(value, ToSQLInterface):
            columns.append(key)
            values.append(value.to_sql())
        else:
            columns.append(key)
            values.append('%s')

    filtered_data_json = {key: value for key, value in data_json.items() if not isinstance(value, ToSQLInterface)}  # noqa501
    return ','.join(columns), ','.join(values), filtered_data_json


def process_update_data_json(data_json):
    set_values = []

    for key, value in data_json.items():
        if isinstance(value, ToSQLInterface):
            set_values.append(f"{key}={value.to_sql()}")
        else:
            set_values.append(f"{key}=%s")

    filtered_data_json = {key: value for key, value in data_json.items() if not isinstance(value, ToSQLInterface)}  # noqa501
    return ', '.join(set_values) + ("," if data_json else ""), filtered_data_json  # noqa501



import re
from typing import Any

import mysql.connector
from dotenv import load_dotenv
from logger_local.Logger import Logger
from logger_local.LoggerComponentEnum import LoggerComponentEnum
from circles_number_generator.number_generator import NumberGenerator

from .connector import Connector
from .utils import validate_none_select_table_name, validate_select_table_name, process_insert_data_json, process_update_data_json  # noqa E402

load_dotenv()

# Constants
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID = 206
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME = 'circles_local_database_python\\generic_crud'
DEVELOPER_EMAIL = 'akiva.s@circ.zone'

# Logger setup
logger = Logger.create_logger(object={
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': DEVELOPER_EMAIL
})


class GenericCRUD:
    """A class that provides generic CRUD functionality"""

    def __init__(self, default_schema_name: str,
                 default_table_name: str = None,
                 default_view_table_name: str = None,
                 default_id_column_name: str = None,
                 connection: Connector = None, is_test_data: bool = False) -> None:
        """Initializes the GenericCRUD class. If connection is not provided, a new connection will be created."""
        logger.start(object={"default_schema_name": default_schema_name,
                             "default_table_name": default_table_name,
                             "id_column_name": default_id_column_name})
        self.schema_name = default_schema_name
        self.connection = connection or Connector.connect(schema_name=default_schema_name)
        self.cursor = self.connection.cursor()
        self.default_column = default_id_column_name
        self.default_table_name = default_table_name
        self.default_view_table_name = default_view_table_name
        self.is_test_data = is_test_data
        logger.end()

    def insert(self, table_name: str = None, data_json: dict = None,
               ignore_duplicate: bool = False) -> int:
        """Inserts a new row into the table and returns the id of the new row or -1 if an error occurred."""
        logger.start(object={"table_name": table_name,
                     "data_json": str(data_json)})
        if ignore_duplicate:
            logger.warn("Using ignore_duplicate, is it really needed?")
        table_name = table_name or self.default_table_name

        if not self._is_ml_table(table_name):
            number = NumberGenerator.get_random_number(schema=self.schema_name, table=table_name)
            data_json["number"] = number
        data_json["is_test_data"] = self.is_test_data

        self._validate_data_json(data_json)
        self._validate_table_name(table_name)
        validate_none_select_table_name(table_name)
        columns, values, data_json = process_insert_data_json(data_json=data_json)
        # We removed the IGNORE from the SQL Statement as we want to return the id of the existing row
        insert_query = "INSERT " + \
                       f"INTO {self.schema_name}.{table_name} ({columns}) " \
                       f"VALUES ({values})"
        try:
            try:
                self.cursor.execute(insert_query, tuple(data_json.values()))
                self.connection.commit()
                inserted_id = self.cursor.lastrowid()
            except mysql.connector.errors.IntegrityError as error:
                if ignore_duplicate:
                    logger.warn("Existing record found, selecting it's id")
                    inserted_id = self._get_existing_duplicate_id(table_name,
                                                                  error)
                else:
                    logger.end()
                    raise error
            logger.end(f"Data inserted successfully with id {inserted_id}.")
            return inserted_id
        except Exception as error:
            logger.exception(self._log_error_message(message="Error inserting data_json",
                                                     sql_statement=insert_query), object=error)
            logger.end()
            raise

    def _get_existing_duplicate_id(self, table_name: str,
                                   error: Exception) -> int:
        pattern = r'Duplicate entry \'(.+?)\' for key \'(.+?)\''
        match = re.search(pattern, str(error))
        if not match:  # a different error
            raise error
        duplicate_value = match.group(1)
        query = """
        SELECT COLUMN_NAME
        FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE
        WHERE TABLE_NAME = %s AND CONSTRAINT_NAME = "PRIMARY"
        """
        self.cursor.execute(query, (table_name,))
        column_name = self.cursor.fetchone()[0]
        if column_name:
            select_query = f"SELECT * FROM {table_name} WHERE {column_name} = %s"
            self.cursor.execute(select_query, (duplicate_value,))
            existing_entry = self.convert_to_dict(self.cursor.fetchone())
            return existing_entry.get(column_name)
        else:  # Column name for constraint not found
            raise error

    # Old name: update
    def update_by_id(self, table_name: str = None, id_column_name: str = None,
                     id_column_value: Any = None, data_json: dict = None,
                     limit: int = 100, order_by: str = "") -> None:
        """Updates data in the table by ID."""
        logger.start(object={"table_name": table_name,
                             "data_json": str(data_json),
                             "id_column_name": id_column_name,
                             "id_column_value": id_column_value,
                             "limit": limit, "order_by": order_by})
        table_name = table_name or self.default_table_name
        id_column_name = id_column_name or self.default_column
        self._validate_data_json(data_json)
        self._validate_table_name(table_name)

        if id_column_name:
            if id_column_value is None:
                where = f"{id_column_name} IS NULL"
                extra_sql_params = None
            else:
                where = f"{id_column_name}=%s"
                extra_sql_params = (id_column_value,)
            self.update_by_where(table_name=table_name, where=where,
                                 data_json=data_json, params=extra_sql_params,
                                 limit=limit, order_by=order_by)
        else:
            message = "Update by id requires an id_column_name"
            logger.error(message)
            logger.end()
            raise Exception(message)

    # Old name: update
    def update_by_where(self, table_name: str = None, where: str = None,
                        params: tuple = None, data_json: dict = None,
                        limit: int = 100, order_by: str = None) -> None:
        """Updates data in the table by WHERE.
        Example:
        "UPDATE table_name SET A=A_val, B=B_val WHERE C=C_val AND D=D_val"
        translates into:
        update_by_where(table_name="table_name",
                        data_json={"A": A_val, "B": B_val},
                        where="C=%s AND D=%s",
                        params=(C_val, D_val)"""
        logger.start(object={"table_name": table_name,
                             "data_json": str(data_json), "where": where,
                             "params": str(params), "limit": limit})
        table_name = table_name or self.default_table_name
        self._validate_data_json(data_json)
        self._validate_table_name(table_name)
        validate_none_select_table_name(table_name)

        set_values = ', '.join(
            f"{k}=%s" for k in data_json.keys()) + ("," if data_json else "")
        set_values, data_json = process_update_data_json(data_json)
        if not where:
            message = "update_by_where requires a 'where'"
            logger.error(message)
            logger.end()
            raise Exception(message)

        update_query = f"UPDATE {self.schema_name}.{table_name} " \
            f"SET {set_values} updated_timestamp=CURRENT_TIMESTAMP() " \
            f"WHERE {where} " + \
            (f"ORDER BY {order_by} " if order_by else "") + \
            f"LIMIT {limit} "
        try:
            params = params or tuple()
            self.cursor.execute(update_query, tuple(
                data_json.values()) + params)
            self.connection.commit()
            logger.end("Data updated successfully.")
        except Exception as e:
            logger.exception(self._log_error_message(message="Error updating data_json",
                                                     sql_statement=update_query), object=e)
            logger.end()
            raise

    def delete_by_id(self, table_name: str = None, id_column_name: str = None,
                     id_column_value: Any = None) -> None:
        """Deletes data from the table by id"""
        # logger, checks etc. are done inside delete_by_where
        id_column_name = id_column_name or self.default_column
        if id_column_name:  # id_column_value can be empty
            if id_column_value is None:
                where = f"{id_column_name} IS NULL"
                params = None
            else:
                where = f"{id_column_name}=%s"
                params = (id_column_value,)
            self.delete_by_where(table_name, where, params)
        else:
            message = "Delete by id requires an id_column_name and id_column_value."
            logger.error(message)
            logger.end()
            raise Exception(message)

    def delete_by_where(self, table_name: str = None, where: str = None,
                        params: tuple = None) -> None:
        """Deletes data from the table by WHERE."""
        logger.start(object={"table_name": table_name,
                     "where": where, "params": str(params)})
        table_name = table_name or self.default_table_name
        self._validate_table_name(table_name)
        if not where:
            message = "delete_by_where requires a 'where'"
            logger.error(message)
            logger.end()
            raise Exception(message)
        update_query = f"UPDATE {self.schema_name}.{table_name} " \
            f"SET end_timestamp=CURRENT_TIMESTAMP() " \
            f"WHERE {where}"
        try:
            self.cursor.execute(update_query, params)
            self.connection.commit()
            logger.end("Deleted successfully.")

        except Exception as e:
            logger.exception(
                self._log_error_message(message="Error while deleting",
                                        sql_statement=update_query), object=e)
            logger.end()
            raise

    # Old name: select_one_by_id
    def select_one_tuple_by_id(self, view_table_name: str = None,
                               select_clause_value: str = "*",
                               id_column_name: str = None,
                               id_column_value: Any = None,
                               order_by: str = "") -> tuple:
        """Selects one row from the table by ID and returns it as a tuple."""
        result = self.select_multi_tuple_by_id(view_table_name,
                                               select_clause_value,
                                               id_column_name, id_column_value,
                                               limit=1, order_by=order_by)
        if result:
            return result[0]
        else:
            return tuple()

    def select_one_dict_by_id(self, view_table_name: str = None,
                              select_clause_value: str = "*",
                              id_column_name: str = None,
                              id_column_value: Any = None,
                              order_by: str = "") -> dict:
        """Selects one row from the table by ID and returns it as a dictionary (column_name: value)"""
        result = self.select_one_tuple_by_id(view_table_name,
                                             select_clause_value,
                                             id_column_name, id_column_value,
                                             order_by=order_by)
        return self.convert_to_dict(result, select_clause_value)

    # Old name: select_one_by_where
    def select_one_tuple_by_where(self, view_table_name: str = None,
                                  select_clause_value: str = "*",
                                  where: str = None, params: tuple = None,
                                  order_by: str = "") -> tuple:
        """Selects one row from the table based on a WHERE clause and returns it as a tuple."""
        result = self.select_multi_tuple_by_where(view_table_name,
                                                  select_clause_value,
                                                  where=where, params=params,
                                                  limit=1, order_by=order_by)
        if result:
            return result[0]
        else:
            return tuple()

    def select_one_dict_by_where(self, view_table_name: str = None,
                                 select_clause_value: str = "*",
                                 where: str = None, params: tuple = None,
                                 order_by: str = "") -> dict:
        """Selects one row from the table based on a WHERE clause and returns it as a dictionary."""
        result = self.select_one_tuple_by_where(view_table_name,
                                                select_clause_value,
                                                where=where, params=params,
                                                order_by=order_by)
        return self.convert_to_dict(result, select_clause_value)

    # Old name: select_multi_by_id
    def select_multi_tuple_by_id(self, view_table_name: str = None,
                                 select_clause_value: str = "*",
                                 id_column_name: str = None,
                                 id_column_value: Any = None,
                                 limit: int = 100, order_by: str = "") -> list:
        """Selects multiple rows from the table by ID and returns them as a
        list of tuples.
        send `id_column_name=''` if you want to select all rows and ignore default column"""
        id_column_name = id_column_name or self.default_column

        if not id_column_name:
            where = None
            params = None
        else:
            if id_column_value is None:
                where = f"{id_column_name} IS NULL"
                params = None
            else:
                where = f"{id_column_name}=%s"
                params = (id_column_value,)
        return self.select_multi_tuple_by_where(view_table_name,
                                                select_clause_value,
                                                where=where, params=params,
                                                limit=limit, order_by=order_by)

    def select_multi_dict_by_id(
            self, view_table_name: str = None, select_clause_value: str = "*",
            id_column_name: str = None, id_column_value: Any = None,
            limit: int = 100, order_by: str = "") -> list:
        """Selects multiple rows from the table by ID and returns them as a list of dictionaries."""
        result = self.select_multi_tuple_by_id(view_table_name,
                                               select_clause_value,
                                               id_column_name, id_column_value,
                                               limit=limit, order_by=order_by)
        return [self.convert_to_dict(row, select_clause_value) for row in result]

    # Old name: select_multi_by_where
    def select_multi_tuple_by_where(self, view_table_name: str = None,
                                    select_clause_value: str = "*",
                                    where: str = None, params: tuple = None,
                                    limit: int = 100,
                                    order_by: str = "") -> list:
        """Selects multiple rows from the table based on a WHERE clause and returns them as a list of tuples."""
        logger.start(object={"default_view_table_name": view_table_name,
                             "select_clause_value": select_clause_value,
                             "where": where, "params": str(params),
                             "limit": limit, "order_by": order_by})
        view_table_name = view_table_name or self.default_view_table_name
        validate_select_table_name(view_table_name)
        select_query = f"SELECT {select_clause_value} " \
            f"FROM {self.schema_name}.{view_table_name} " + \
            (f"WHERE {where} " if where else "") + \
            (f"ORDER BY {order_by} " if order_by else "") + \
            f"LIMIT {limit}"
        try:
            self.cursor.execute(select_query, params)
            result = self.cursor.fetchall()
            logger.end("Data selected successfully.",
                       object={"result": str(result)})
            return result
        except Exception as e:
            logger.exception(self._log_error_message(message="Error selecting data_json",
                                                     sql_statement=select_query), object=e)
            logger.end()
            raise

    def select_multi_dict_by_where(
            self, view_table_name: str, select_clause_value: str = "*",
            where: str = None, params: tuple = None,
            limit: int = 100, order_by: str = "") -> list:
        """Selects multiple rows from the table based on a WHERE clause and returns them as a list of dictionaries."""
        result = self.select_multi_tuple_by_where(view_table_name,
                                                  select_clause_value,
                                                  where=where, params=params,
                                                  limit=limit,
                                                  order_by=order_by)
        return [self.convert_to_dict(row, select_clause_value) for row in result]

    # helper functions:

    def switch_db(self, new_database: str) -> None:
        """Switches the database to the given database name."""
        logger.start(object={"default_schema_name": new_database})
        self.connection.set_schema(new_database)
        self.schema_name = new_database
        logger.end("Schema set successfully.")

    def convert_to_dict(self, row: tuple, select_clause_value: str = "*") -> dict:
        """Returns a dictionary of the column names and their values."""
        if select_clause_value == "*":
            column_names = [col[0] for col in self.cursor.description()]
        else:
            column_names = [x.strip() for x in select_clause_value.split(",")]
        return dict(zip(column_names, row or tuple()))

    @staticmethod
    def _validate_table_name(table_name: str) -> None:
        """Validates the table name."""
        if not table_name:
            message = "Table name is required."
            logger.error(message)
            logger.end()
            raise Exception(message)

    @staticmethod
    def _validate_data_json(data_json: dict) -> None:
        """Validates the json data."""
        if not data_json:
            message = "Json data is required."
            logger.error(message)
            logger.end()
            raise Exception(message)

    def set_schema(self, schema_name: str):
        """Sets the schema to the default schema."""
        logger.start()
        self.connection.set_schema(schema_name)
        self.schema_name = schema_name
        logger.end()

    def close(self) -> None:
        """Closes the connection to the database."""
        logger.start()
        self.connection.close()
        logger.end()

    @staticmethod
    def _log_error_message(message: str, sql_statement: str) -> str:
        return f"{message} - SQL statement: {sql_statement}"

    def _is_ml_table(self, table_name: str) -> bool:
        """Returns True if the table is an ML table, False otherwise."""
        return table_name.endswith("_ml_table")



from abc import ABC, abstractmethod


class ToSQLInterface(ABC):
    """
    An interface for objects that represent structures to be
    inserted into a database.

    Subclasses must implement the `to_sql` method, which should return a string
    representing the SQL representation of the structure.

    Example:
    --------
    class Point(ToSQLInterface):
        def __init__(self, longitude, latitude):
            self.longitude = longitude
            self.latitude = latitude

        def to_sql(self):
            return f"POINT({self.longitude}, {self.latitude})"
    """

    @abstractmethod
    def to_sql(self) -> str:
        pass


class Point(ToSQLInterface):
    def __init__(self, longitude, latitude):
        self.longitude = longitude
        self.latitude = latitude

    def to_sql(self):
        return f"POINT ({self.longitude}, {self.latitude})"

    @staticmethod
    def create_select_stmt(column_name: str):
        return f"ST_X({column_name}), ST_Y({column_name})"

    def __eq__(self, other_point):
        return (isinstance(other_point, Point) and
                self.longitude == other_point.longitude and
                self.latitude == other_point.latitude)

    def __repr__(self):
        return f"Point(longitude={self.longitude}, latitude={self.latitude})"


class Now(ToSQLInterface):
    """
    Represents the current time in SQL.
    """

    def to_sql(self):
        return "NOW()"



from typing import Any
import re

from logger_local.Logger import Logger
from logger_local.LoggerComponentEnum import LoggerComponentEnum
from language_local.lang_code import LangCode
from user_context_remote.user_context import UserContext

from .generic_crud import GenericCRUD
from .connector import Connector
from .utils import validate_select_table_name

# Constants
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID = 206
DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME = 'circles_local_database_python\\generic_crud_ml'
DEVELOPER_EMAIL = 'tal.g@circ.zone'
DEFAULT_LIMIT = 100

user_context = UserContext()
user = user_context.login_using_user_identification_and_password()

# Logger setup
logger = Logger.create_logger(object={
    'component_id': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_GENERIC_CRUD_ML_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': DEVELOPER_EMAIL
})
'''
city_id = cityMl.add_value( None, ENGLISH, "Jerusalem"...)
cityMl.add_value( city_id, HEBREW, "ירושלים"...)
city_name_english = cityML.get_value( city_id, ENGLISH)
get_id_by_name( name: str, lang_code: LangCode = None) -> int
'''

TEST_TABLE_NAME = 'test_mysql_table'
TEST_ML_TABLE_NAME = 'test_mysql_ml_table'


class GenericCRUDML(GenericCRUD):
    """A class that provides generic CRUD functionality for tables with multi-language support."""

    def __init__(self, default_schema_name: str, default_table_name: str = None,
                 default_id_column_name: str = None, connection: Connector = None,
                 is_test_data: bool = False) -> None:
        """Initializes the GenericCRUDML class. If connection is not provided,
        a new connection will be created."""
        logger.start(object={"default_schema_name": default_schema_name,
                             "default_table_name": default_table_name,
                             "id_column_name": default_id_column_name})

        if default_table_name is not None:
            self.default_ml_table_name = self._create_ml_table_name(default_table_name)
            self.default_view_table_name = self._create_view_name(default_table_name)
        else:
            self.default_ml_table_name = None
            self.default_view_table_name = None
        #     self.default_ml_table_name = default_ml_table_name or re.sub(r'(_table)$', '_ml\\1', default_table_name)
        logger.info(object={"default_ml_table_name": self.default_ml_table_name,
                            "default_view_table_name": self.default_view_table_name})
        super().__init__(default_schema_name=default_schema_name, default_table_name=default_table_name,
                         default_view_table_name=self.default_view_table_name, default_id_column_name=default_id_column_name,
                         connection=connection, is_test_data=is_test_data)

        logger.end()

    def sql_in_list_by_entity_list_id(self, schema_name: str, entity_name: str, entity_list_id: int) -> str:
        """Example: select group_id from group.group_list_member_table WHERE group_list_id=1"""
        old_schema_name = self.schema_name
        self.set_schema(schema_name)
        ids = self.select_multi_dict_by_id(view_table_name=f"{entity_name}_list_member_view",
                                           select_clause_value=f"{entity_name}_id",
                                           id_column_name=f"{entity_name}_list_id",
                                           id_column_value=entity_list_id)
        ids = ids or [{f"{entity_name}_id": -1}]
        result = f" IN ({','.join([str(_id[f'{entity_name}_id']) for _id in ids])})"
        self.set_schema(old_schema_name)
        return result

    def add_value(self, table_id: int = None, lang_code: LangCode = None,
                  is_main: int = 0, data_json: dict = None, data_ml_json: dict = None,
                  table_name: str = None, ml_table_name: str = None,
                  view_name: str = None) -> tuple[int, int]:
        logger.start(object={"table_id": table_id, "lang_code": lang_code, "data_json": str(data_json),
                             "data_ml_json": str(data_ml_json), "table_name": table_name,
                             "ml_table_name": ml_table_name, "view_name": view_name})

        lang_code_str = self._get_lang_code_str(lang_code)
        table_name = table_name or self.default_table_name
        ml_table_name = ml_table_name or self.default_ml_table_name
        view_name = view_name or self.default_view_table_name

        # if this is the first insert of this data, is_main should be 1
        if table_id is None:
            is_main = 1
        elif is_main == 1:
            self._update_old_main_value_to_zero(table_id, table_name)

        # id is the id value of the row in the table_name table
        table_id = table_id or self.insert(data_json=data_json, ignore_duplicate=True)
        if table_id is None:    # TODO: delete
            logger.error("Error inserting data_json", object={"data_json": str(data_json)})
            logger.end()
            return None

        if data_ml_json is None or data_ml_json == {}:
            logger.end(object={"table_id": table_id})
            return table_id, None

        id_column_name = re.sub(r'(_table)$', '_id', table_name)
        data_ml_json[id_column_name] = table_id
        data_ml_json["lang_code"] = lang_code_str
        data_ml_json["is_main"] = is_main

        # ml_id is the id value of the row in the ml_table_name table
        ml_table_id = self.insert(data_json=data_ml_json, table_name=ml_table_name, ignore_duplicate=True)

        logger.end(object={"table_id": table_id, "ml_table_id": ml_table_id})
        return table_id, ml_table_id

    def get_values_tuple(self, table_id: int, lang_code: LangCode = None,
                         id_column_name: str = None) -> tuple:
        logger.start(object={"table_id": table_id, "lang_code": lang_code})

        lang_code_str = self._get_lang_code_str(lang_code)
        id_column_name = self._generate_id_column_name(id_column_name, self.default_table_name)
        result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                                params=(table_id, lang_code_str))
        logger.end(object={"result": result})
        return result

    def get_values_dict(self, table_id: int, lang_code: LangCode = None,
                        id_column_name: str = None) -> dict:
        logger.start(object={"table_id": table_id, "lang_code": lang_code})

        lang_code_str = self._get_lang_code_str(lang_code)
        id_column_name = self._generate_id_column_name(id_column_name, self.default_table_name)
        result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND lang_code=%s",
                                               params=(table_id, lang_code_str))
        logger.end(object={"result": result})
        return result

    def get_main_values_tuple(self, table_id: int, id_column_name: str = None) -> tuple:
        logger.start(object={"table_id": table_id})

        id_column_name = self._generate_id_column_name(id_column_name, self.default_table_name)
        result = self.select_one_tuple_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                                params=(table_id,))
        logger.end(object={"result": result})
        return result

    def get_main_values_dict(self, table_id: int, id_column_name: str = None) -> dict:
        logger.start(object={"table_id": table_id, "id_column_name": id_column_name})

        id_column_name = self._generate_id_column_name(id_column_name, self.default_table_name)
        result = self.select_one_dict_by_where(where=f"{id_column_name}=%s AND is_main=1",
                                               params=(table_id,))
        logger.end(object={"result": result})
        return result

    def get_id_by_name(self, name: str, lang_code: LangCode = None,
                       id_column_name: str = None) -> int:
        logger.start(object={"name": name, "lang_code": lang_code})

        lang_code_str = self._get_lang_code_str(lang_code)
        id_column_name = self._generate_id_column_name(id_column_name, self.default_table_name)
        try:
            result = self.select_one_tuple_by_where(select_clause_value=id_column_name,
                                                    where="title=%s AND lang_code=%s",
                                                    params=(name, lang_code_str))
        except Exception:
            logger.warn("select_one_tuple_by_where was not successful for name and lang_code",
                        object={"title": name, "lang_code": lang_code})
            try:
                # this call is not tested because test_mysql_ml_table does not have a `name` column
                result = self.select_one_tuple_by_where(select_clause_value=id_column_name,
                                                        where="`name`=%s AND lang_code=%s",
                                                        params=(name, lang_code_str))
            except Exception:
                logger.warn("select_one_tuple_by_where was not successful for name and lang_code",
                            object={"name": name, "lang_code": lang_code})
                logger.end()
                raise
        logger.end(object={"result": result})
        if result:
            return result[0]
        else:
            return None

    def get_ml_id_by_name(self, name: str, lang_code: LangCode = None,
                          id_column_name: str = None) -> int:
        logger.start(object={"name": name, "lang_code": lang_code})

        lang_code_str = self._get_lang_code_str(lang_code)
        ml_id_column_name = self._generate_id_column_name(id_column_name, self.default_ml_table_name)
        try:
            result = self.select_one_tuple_by_where(select_clause_value=ml_id_column_name,
                                                    where="title=%s AND lang_code=%s",
                                                    params=(name, lang_code_str))
        except Exception:
            logger.warn("select_one_tuple_by_where was not successful for name and lang_code",
                        object={"title": name, "lang_code": lang_code})
            try:
                # this call is not tested because test_mysql_ml_view does not have a `name` column
                result = self.select_one_tuple_by_where(select_clause_value=ml_id_column_name,
                                                        where="`name`=%s AND lang_code=%s",
                                                        params=(name, lang_code_str))
            except Exception:
                logger.warn("select_one_tuple_by_where was not successful for name and lang_code",
                            object={"name": name, "lang_code": lang_code})
                logger.end()
                raise
        logger.end(object={"result": result})
        return result[0]

    def delete_by_name(self, name: str,
                       lang_code: LangCode = None) -> None:
        logger.start(object={"name": name, "lang_code": lang_code})

        lang_code_str = self._get_lang_code_str(lang_code)
        try:
            self.delete_by_where(table_name=TEST_ML_TABLE_NAME, where="title=%s AND lang_code=%s",
                                 params=(name, lang_code_str))
        except Exception:
            logger.warn("delete_by_where was not successful for name and lang_code",
                        object={"title": name, "lang_code": lang_code})
            try:
                # this call is not tested because test_mysql_ml_table does not have a `name` column
                self.delete_by_where(table_name=TEST_ML_TABLE_NAME,
                                     where="`name`=%s AND lang_code=%s", params=(name, lang_code_str))
            except Exception:
                logger.warn("delete_by_where was not successful for name and lang_code",
                            object={"name": name, "lang_code": lang_code})
                logger.end()
                raise
        logger.end()

    def _create_ml_table_name(self, table_name: str) -> str:
        ml_table_name = re.sub(r'(_table)$', '_ml_table', table_name)
        return ml_table_name

    def _create_view_name(self, table_name: str) -> str:
        view_name = re.sub(r'(_table)$', '_ml_view', table_name)
        return view_name

    def _generate_id_column_name(self, id_column_name: str, table_name: str) -> str:
        try:
            id_column_name = id_column_name or re.sub(r'(_table)$', '_id', table_name)
        except Exception as exception:
            message = "id_column_name was not provided and could not be generated from default_table_name"
            logger.exception(message, exception)
            logger.end()
            raise Exception(message)
        return id_column_name

    # Change the old row with is_main=1 to is_main=0
    def _update_old_main_value_to_zero(self, table_id: int, table_name: str) -> None:
        logger.start(object={"table_id": table_id, "table_name": table_name})
        data_ml_json = {"is_main": 0}
        id_column_name = self._generate_id_column_name(None, table_name)
        try:
            self.update_by_where(table_name=self.default_ml_table_name,
                                 data_json=data_ml_json,
                                 where=f"{id_column_name}=%s AND is_main=1",
                                 params=(table_id,))
        except Exception as exception:
            message = "Error updating old main"
            logger.exception(message, exception)
            logger.end()
            raise Exception(message)
        logger.end()

    def _get_lang_code_str(self, lang_code: LangCode) -> str:
        if lang_code is None:
            lang_code_str = user_context.get_effective_profile_preferred_lang_code()
        else:
            lang_code_str = lang_code.value
        return lang_code_str

    # The follwing methods may be irrelevant since GenericCRUD has equivalent methods
    # TODO: We may want to remove these following methods

    # This method returns the value in a selected column by a condition, the condition can be none or a value  # noqa501
    # for a specific column
    def select_one_by_value_with_none_option(
            self, condition_column_name, condition_column_value,
            view_table_name: str = None,
            select_column_name: str = None) -> Any:
        view_table_name = view_table_name or self.default_view_table_name
        # Returns the row id if select_column_name is None
        select_column_name = select_column_name or self.default_column
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""  # noqa501
        logger.start(
            object={"table_name": view_table_name,
                    "select_column_name": select_column_name,
                    "condition_column_name": condition_column_name,
                    "condition_column_value": condition_column_value})
        logger.warn("this method is deprecated, use select_one_tuple_by_id instead")
        validate_select_table_name(view_table_name)
        select_query = f"SELECT {select_column_name} " f"FROM {self.schema_name}.{view_table_name} " + (
            f"WHERE `{condition_column_name}` = \"{condition_column_value}\" " if (condition_column_name and
                                                                                   condition_column_value)
            else f"WHERE `{condition_column_name}` IS NULL ")
        try:
            self.cursor.execute(select_query)
            result = self.cursor.fetchall()
            # Extract the first element from the first tuple in the result
            result = result[0][0] if result else None
            logger.end(object={"result": str(result)})
            return result
        except Exception as e:
            logger.exception(self._log_error_message(message="Error selecting data_json",  # noqa501
                                                     sql_statement=select_query), object=e)  # noqa501
            logger.end()
            raise

    def select_multi_tuple_by_value_with_none_option(self,
                                                     condition_column_name,
                                                     condition_column_value,
                                                     view_table_name: str = None, select_clause_value: str = "*",  # noqa501
                                                     limit: int = DEFAULT_LIMIT, order_by: str = "") -> list:  # noqa501
        view_table_name = view_table_name or self.default_view_table_name
        # Returns the row id if select_column_name is None
        select_column_name = f"`{select_clause_value}`" if select_clause_value != "*" else "*"
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""
        logger.start(
            object={"table_name": view_table_name,
                    "select_column_name": select_column_name,
                    "condition_column_name": condition_column_name,
                    "condition_column_value": condition_column_value})
        logger.warn("this method is deprecated, use select_multi_tuple_by_id instead")
        validate_select_table_name(view_table_name)
        select_query = f"SELECT {select_column_name} " f"FROM {self.schema_name}.{view_table_name} " + (
            f"WHERE `{condition_column_name}` = \"{condition_column_value}\" " if (condition_column_name and
                                                                                   condition_column_value)
            else f"WHERE {condition_column_name} IS NULL ") + (
                           f"ORDER BY {order_by} " if order_by else "") + f"LIMIT {limit}"  # noqa501
        try:
            self.cursor.execute(select_query)
            result = self.cursor.fetchall()
            logger.end("Data selected successfully.",
                       object={"result": str(result)})
            return result
        except Exception as e:
            logger.exception(self._log_error_message(message="Error selecting data_json",  # noqa501
                                                     sql_statement=select_query), object=e)  # noqa501
            logger.end()
            raise

    def select_multi_dict_by_value_with_none_option(self,
                                                    condition_column_name,
                                                    condition_column_value,
                                                    view_table_name: str = None, select_clause_value: str = "*",  # noqa501
                                                    limit: int = DEFAULT_LIMIT, order_by: str = "") -> list:  # noqa501
        logger.warn("this method is deprecated, use select_multi_tuple_by_id instead")
        """Selects a column from the table based on a WHERE clause and returns it as a list of dictionaries."""  # noqa501
        result = self.select_multi_tuple_by_value_with_none_option(condition_column_name, condition_column_value,  # noqa501
                                                                   view_table_name=view_table_name,  # noqa501
                                                                   select_clause_value=select_clause_value,  # noqa501
                                                                   limit=limit, order_by=order_by)  # noqa501  
        return [self.convert_to_dict(row, select_clause_value) for row in result]  # noqa501



import datetime
from dotenv import load_dotenv
load_dotenv()
from .utils import validate_none_select_table_name  # noqa E402
from .generic_crud import GenericCRUD  # noqa E402
from logger_local.LoggerComponentEnum import LoggerComponentEnum  # noqa E402
from logger_local.Logger import Logger  # noqa E402


# Constants
# TODO make the component id for the generic mapping component if needed
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID = 206
DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME = 'circles_local_database_python\\generic_mapping'
DEVELOPER_EMAIL = 'sahar.g@circ.zone'

# Logger setup
logger = Logger.create_logger(object={
    'component_id': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_ID,
    'component_name': DATABASE_MYSQL_PYTHON_GENERIC_CRUD_COMPONENT_NAME,
    'component_category': LoggerComponentEnum.ComponentCategory.Code.value,
    'developer_email': DEVELOPER_EMAIL
})


class GenericMapping(GenericCRUD):

    def __init__(self, default_schema_name: str = None, default_table_name: str = None, default_view_table_name: str = None,
                 default_id_column_name: str = None, default_entity_name1: str = None, default_entity_name2: str = None,
                 is_test_data: bool = False):
        super().__init__(default_schema_name=default_schema_name,
                         default_table_name=default_table_name, default_view_table_name=default_view_table_name,
                         default_id_column_name=default_id_column_name, is_test_data=is_test_data)
        self.default_schema_name = default_schema_name
        self.default_table_name = default_table_name
        self.default_view_table_name = default_view_table_name
        self.default_id_column_name = default_id_column_name
        self.default_entity_name1 = default_entity_name1
        self.default_entity_name2 = default_entity_name2

    def insert_mapping(self, entity_name1: str, entity_name2: str, entity_id1: int, entity_id2: int) -> int:
        """Inserts a new link between two entities and returns the id of the
          new row or -1 if an error occurred.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :return: The id of the new row or -1 if an error occurred.
        """
        logger.start(object={"entity_name1": entity_name1, "entity_name2": entity_name2, "entity_id1": entity_id1,
                             "entity_id2": entity_id2})

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        combined_table_name = f"{entity_name1}_{entity_name2}_table"
        data_json = {
            f"{entity_name1}_id": entity_id1,
            f"{entity_name2}_id": entity_id2}
        self._validate_data_json(data_json)
        self._validate_table_name(entity_name1)
        validate_none_select_table_name(combined_table_name)
        columns = ','.join(data_json.keys())
        values = ','.join(['%s' for _ in data_json])
        insert_query = f"INSERT " \
                       f"INTO {self.schema_name}.{combined_table_name}({columns}) " \
                       f"VALUES ({values})"
        params = tuple(data_json.values())
        try:
            self.cursor.execute(insert_query, params)
            self.connection.commit()
            link_id = self.cursor.lastrowid()
            logger.end("Data inserted successfully.",
                       object={"link_id": link_id})
            return link_id
        except Exception as error:
            logger.exception(self._log_error_message(message="Error inserting data_json",
                                                     sql_statement=insert_query), object=error)
            logger.end()
            raise

    def delete_mapping_by_id(self, entity_name1: str, entity_name2: str, entity_id1: int, entity_id2: int) -> None:
        """ Deletes a link between two entities.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :return: None
        """
        logger.start(object={"entity_name1": entity_name1, "entity_name2": entity_name2, "entity_id1": entity_id1,
                             "entity_id2": entity_id2})

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        combined_table_name = f"{entity_name1}_{entity_name2}_table"

        where = f"{entity_name1}_id=%s AND {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        self._validate_table_name(combined_table_name)
        validate_none_select_table_name(combined_table_name)
        update_query = f"UPDATE {self.schema_name}.{combined_table_name} " \
                       f"SET end_timestamp=CURRENT_TIMESTAMP() " \
                       f"WHERE {where}"
        try:
            self.cursor.execute(update_query, params)
            self.connection.commit()
            logger.end("Deleted successfully.")

        except Exception as e:
            logger.exception(
                self._log_error_message(message="Error while deleting", sql_statement=update_query), object=e)
            logger.end()
            raise

    def select_multi_mapping_tupel_by_id(self, entity_name1: str,
                                         entity_name2: str, entity_id1: int,
                                         entity_id2: int, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """
        logger.start(object={"entity_name1": entity_name1,
                             "entity_name2": entity_name2,
                             "entity_id1": entity_id1,
                             "entity_id2": entity_id2,
                             "select_clause_value": select_clause_value})

        entity_name1 = entity_name1 or self.default_entity_name1
        entity_name2 = entity_name2 or self.default_entity_name2
        combined_table_name = f"{entity_name1}_{entity_name2}_table"
        where = f"{entity_name1}_id=%s OR {entity_name2}_id=%s"
        params = (entity_id1, entity_id2)
        self._validate_table_name(combined_table_name)
        validate_none_select_table_name(combined_table_name)
        select_query = f"SELECT {select_clause_value} " \
                       f"FROM {self.schema_name}.{combined_table_name} " \
                       f"WHERE {where}"
        try:
            self.cursor.execute(select_query, params)
            result = self.cursor.fetchall()
            result = [tuple(str(value) if isinstance(
                value, datetime.datetime) else value for value in row) for row in result]
            logger.end("Data selected successfully.",
                       object={"result": result})
            return result
        except Exception as e:
            logger.exception(
                self._log_error_message(message="Error selecting data",
                                        sql_statement=select_query), object=e)
            logger.end()
            raise

    def select_multi_mapping_dict_by_id(self, entity_name1: str,
                                        entity_name2: str, entity_id1: int,
                                        entity_id2: int, select_clause_value: str = "*") -> list:
        """Selects a row from the database by id.
        :param entity_name1: The name of the first entity's table.
        :param entity_name2: The name of the second entity's table.
        :param entity_id1: The id of the first entity.
        :param entity_id2: The id of the second entity.
        :param select_clause_value: The columns to select.
        :return: A list of dictionaries representing the rows.
        """
        logger.start(object={"entity_name1": entity_name1,
                             "entity_name2": entity_name2,
                             "entity_id1": entity_id1,
                             "entity_id2": entity_id2,
                             "select_clause_value": select_clause_value})

        try:
            result = self.select_multi_mapping_tupel_by_id(entity_name1=entity_name1, entity_name2=entity_name2,
                                                           entity_id1=entity_id1, entity_id2=entity_id2,
                                                           select_clause_value=select_clause_value)
            result = [dict(
                zip([column[0] for column in self.cursor.description()], row)) for row in result]
        except Exception as e:
            logger.exception("Error selecting data", object=e)
            logger.end()
            raise
        logger.end("Data selected successfully.",
                   object={"result": result})
        return result
